#include "APF.h"

#include <queue>

Grid *grids = NULL;
unsigned long idcounter_grids = 0;

Grid::Grid(unsigned long _width, unsigned long _height) {
	width = _width;
	height = _height;
	data.resize(width*height);
	prev = NULL;
	next = grids;
	if(next!=NULL) next->prev = this;
	grids = this;
	id = ++idcounter_grids;
}

Grid::~Grid() {
	Passage *np;
	for(Passage *passage = passages; passage!=NULL; passage = np) {
		np = passage->next;
		if(passage->point1.p.grid==this || passage->point2.p.grid==this) {
			delete passage;
		}
	}
	if(prev!=NULL) prev->next = next;
	else grids = next;
	if(next!=NULL) next->prev = prev;
}

void Grid::Clear() {
	Passage *np;
	for(Passage *passage = passages; passage!=NULL; passage = np) {
		np = passage->next;
		if(passage->point1.p.grid==this || passage->point2.p.grid==this) {
			delete passage;
		}
	}
	data.clear();
	data.resize(width*height);
}

Grid* Grid::Find(unsigned long id) {
	for(Grid *grid = grids; grid!=NULL; grid = grid->next) {
		if(grid->id==id) return grid;
	}
	return NULL;
}

struct PathPoint : public GridPoint {
	bool diagonal;
	unsigned long cost;
	inline PathPoint() {};
	inline PathPoint(const GridPoint& other, bool _diagonal, unsigned long extracost) {
		grid = other.grid;
		x = other.x;
		y = other.y;
		cost = grid->GetCell(x, y).cost+extracost;
		diagonal = _diagonal;
	};
	inline PathPoint(const PathPoint& other) {
		grid = other.grid;
		x = other.x;
		y = other.y;
		cost = other.cost;
		diagonal = other.diagonal;
	};
	inline PathPoint(Grid* _grid, unsigned long _x, unsigned long _y, bool _diagonal, unsigned long _cost) {
		grid = _grid;
		x = _x;
		y = _y;
		diagonal = _diagonal;
		cost = _cost;
	}
	inline bool operator<(const PathPoint& other) const {
		return (cost>other.cost); // lower cost = higher priority
	}
};

bool Grid::FindPath(
	const std::vector<GridPoint>& startpoints,
	const std::vector<GridPoint>& endpoints,
	std::vector<unsigned char>* path,
	GridPoint* startpoint, GridPoint* endpoint, unsigned long* cost,
	unsigned long cost_normal, unsigned long cost_diagonal) {
	// The simplest way to find a path is to go flood-fill from the endpoint
	// until the startpoint is reached. But if there are more endpoints than
	// startpoints it is faster to do it the other way around.
	// The directions are like the keys on the numeric keyboard:
	// 8 = up, 2 = down, 4 = left, 6 = right, ...
	// this is also very easy to reverse:
	// reversed = 10-original;
	// 0 = not visited, 5 = endpoint, 10 = startpoint
	for(Grid *grid = grids; grid!=NULL; grid = grid->next) {
		Cell *data = &(grid->data.front());
		unsigned long c = grid->width*grid->height;
		for(unsigned long i = 0; i<c; ++i) {
			data[i].temp = 0;
		}
	}
	std::priority_queue<PathPoint> points;
	bool default_diagonal;
	unsigned long default_cost;
	if(cost_diagonal==0 || cost_normal<=cost_diagonal) {
		default_diagonal = false;
		default_cost = cost_normal;
	} else {
		default_diagonal = true;
		default_cost = cost_diagonal;
	}
	for(unsigned long i = 0; i<startpoints.size(); ++i) {
		startpoints[i].grid->GetCell(startpoints[i].x, startpoints[i].y).temp = 5; // this has to be done first so the endpoints can overwrite it
	}
	for(unsigned long i = 0; i<endpoints.size(); ++i) {
		points.push(PathPoint(endpoints[i], default_diagonal, default_cost));
		endpoints[i].grid->GetCell(endpoints[i].x, endpoints[i].y).temp = 10;
	}
	PathPoint found;
	for(; ; ) {
		if(points.size()==0) {
			return false;
		}
		PathPoint p = points.top();
		points.pop();
		Grid *g = p.grid;
		if(p.diagonal) {
			
			// diagonal
			if(p.x>0) {
				if(p.y>0) { // go up-left
					if(g->GetCell(p.x-1, p.y-1).bottomrightcorner==0) {
						Cell &c = g->GetCell(p.x-1, p.y-1);
						if(c.temp==0) {
							points.push(PathPoint(g, p.x-1, p.y-1, default_diagonal, p.cost+default_cost+c.cost));
							c.temp = 3;
						} else if(c.temp==5) {
							c.temp = 3;
							found = PathPoint(g, p.x-1, p.y-1, default_diagonal, p.cost+c.cost);
							break;
						}
					}
				}
				if(p.y<g->height-1) { // go down-left
					if(g->GetCell(p.x-1, p.y).bottomrightcorner==0) {
						Cell &c = g->GetCell(p.x-1, p.y+1);
						if(c.temp==0) {
							points.push(PathPoint(g, p.x-1, p.y+1, default_diagonal, p.cost+default_cost+c.cost));
							c.temp = 9;
						} else if(c.temp==5) {
							c.temp = 9;
							found = PathPoint(g, p.x-1, p.y+1, default_diagonal, p.cost+c.cost);
							break;
						}
					}
				}
			}
			if(p.x<g->width-1) {
				if(p.y>0) { // go up-right
					if(g->GetCell(p.x, p.y-1).bottomrightcorner==0) {
						Cell &c = g->GetCell(p.x+1, p.y-1);
						if(c.temp==0) {
							points.push(PathPoint(g, p.x+1, p.y-1, default_diagonal, p.cost+default_cost+c.cost));
							c.temp = 1;
						} else if(c.temp==5) {
							c.temp = 1;
							found = PathPoint(g, p.x+1, p.y-1, default_diagonal, p.cost+c.cost);
							break;
						}
					}
				}
				if(p.y<g->height-1) { // go down-right
					if(g->GetCell(p.x, p.y).bottomrightcorner==0) {
						Cell &c = g->GetCell(p.x+1, p.y+1);
						if(c.temp==0) {
							points.push(PathPoint(g, p.x+1, p.y+1, default_diagonal, p.cost+cost_diagonal+c.cost));
							c.temp = 7;
						} else if(c.temp==5) {
							c.temp = 7;
							found = PathPoint(g, p.x+1, p.y+1, default_diagonal, p.cost+c.cost);
							break;
						}
					}
				}
			}
			
			if(cost_normal>cost_diagonal) {
				p.diagonal = false;
				p.cost += cost_normal-cost_diagonal;
				points.push(p);
			}
			
		} else {
			
			// normal
			PassagePoint **passage = p.grid->GetCell(p.x, p.y).passage;
			PassagePoint *pp;
			if((pp = passage[1])!=NULL) { // go left passage
				Cell &c = pp->p.grid->GetCell(pp->p.x, pp->p.y);
				if(c.temp==0) {
					points.push(PathPoint(pp->p.grid, pp->p.x, pp->p.y, default_diagonal, p.cost+default_cost+c.cost+pp->cost));
					c.temp = 6;
				} else if(c.temp==5) {
					c.temp = 6;
					found = PathPoint(pp->p.grid, pp->p.x, pp->p.y, default_diagonal, p.cost+c.cost+pp->cost);
					break;
				}
			} else if(p.x>0) { // go left
				if(g->GetCell(p.x-1, p.y).rightborder==0 && g->GetCell(p.x-1, p.y).passage[2]==NULL) {
					Cell &c = g->GetCell(p.x-1, p.y);
					if(c.temp==0) {
						points.push(PathPoint(g, p.x-1, p.y, default_diagonal, p.cost+default_cost+c.cost));
						c.temp = 6;
					} else if(c.temp==5) {
						c.temp = 6;
						found = PathPoint(g, p.x-1, p.y, default_diagonal, p.cost+c.cost);
						break;
					}
				}
			}
			if((pp = passage[2])!=NULL) { // go right passage
				Cell &c = pp->p.grid->GetCell(pp->p.x, pp->p.y);
				if(c.temp==0) {
					points.push(PathPoint(pp->p.grid, pp->p.x, pp->p.y, default_diagonal, p.cost+default_cost+c.cost+pp->cost));
					c.temp = 4;
				} else if(c.temp==5) {
					c.temp = 4;
					found = PathPoint(pp->p.grid, pp->p.x, pp->p.y, default_diagonal, p.cost+c.cost+pp->cost);
					break;
				}
			} else if(p.x<g->width-1) { // go right
				if(g->GetCell(p.x, p.y).rightborder==0 && g->GetCell(p.x+1, p.y).passage[1]==NULL) {
					Cell &c = g->GetCell(p.x+1, p.y);
					if(c.temp==0) {
						points.push(PathPoint(g, p.x+1, p.y, default_diagonal, p.cost+default_cost+c.cost));
						c.temp = 4;
					} else if(c.temp==5) {
						c.temp = 4;
						found = PathPoint(g, p.x+1, p.y, default_diagonal, p.cost+c.cost);
						break;
					}
				}
			}
			if((pp = passage[3])!=NULL) { // go up passage
				Cell &c = pp->p.grid->GetCell(pp->p.x, pp->p.y);
				if(c.temp==0) {
					points.push(PathPoint(pp->p.grid, pp->p.x, pp->p.y, default_diagonal, p.cost+default_cost+c.cost+pp->cost));
					c.temp = 2;
				} else if(c.temp==5) {
					c.temp = 2;
					found = PathPoint(pp->p.grid, pp->p.x, pp->p.y, default_diagonal, p.cost+c.cost+pp->cost);
					break;
				}
			} else if(p.y>0) { // go up
				if(g->GetCell(p.x, p.y-1).bottomborder==0 && g->GetCell(p.x, p.y-1).passage[0]==NULL) {
					Cell &c = g->GetCell(p.x, p.y-1);
					if(c.temp==0) {
						points.push(PathPoint(g, p.x, p.y-1, default_diagonal, p.cost+default_cost+c.cost));
						c.temp = 2;
					} else if(c.temp==5) {
						c.temp = 2;
						found = PathPoint(g, p.x, p.y-1, default_diagonal, p.cost+c.cost);
						break;
					}
				}
			}
			if((pp = passage[0])!=NULL) { // go down passage
				Cell &c = pp->p.grid->GetCell(pp->p.x, pp->p.y);
				if(c.temp==0) {
					points.push(PathPoint(pp->p.grid, pp->p.x, pp->p.y, default_diagonal, p.cost+default_cost+c.cost+pp->cost));
					c.temp = 8;
				} else if(c.temp==5) {
					c.temp = 8;
					found = PathPoint(pp->p.grid, pp->p.x, pp->p.y, default_diagonal, p.cost+c.cost+pp->cost);
					break;
				}
			} else if(p.y<g->height-1) { // go down
				if(g->GetCell(p.x, p.y).bottomborder==0 && g->GetCell(p.x, p.y+1).passage[3]==NULL) {
					Cell &c = g->GetCell(p.x, p.y+1);
					if(c.temp==0) {
						points.push(PathPoint(g, p.x, p.y+1, default_diagonal, p.cost+default_cost+c.cost));
						c.temp = 8;
					} else if(c.temp==5) {
						c.temp = 8;
						found = PathPoint(g, p.x, p.y+1, default_diagonal, p.cost+c.cost);
						break;
					}
				}
			}
			
			if(cost_diagonal>=cost_normal) {
				p.diagonal = true;
				p.cost += cost_diagonal-cost_normal;
				points.push(p);
			}
			
		}
	}
	if(cost!=NULL) *cost = found.cost;
	if(startpoint!=NULL) *startpoint = found;
	(*path).clear();
	for(; ; ) {
		Cell &c = found.grid->GetCell(found.x, found.y);
		unsigned short dir = c.temp;
		if(dir==10) break;
		(*path).push_back((unsigned char)(dir));
		do {
			if((dir&1)==0 && dir>=2 && dir<=8) {
				unsigned short i = dir/2-1;
				if(c.passage[i]!=NULL) {
					found.grid = c.passage[i]->p.grid;
					found.x = c.passage[i]->p.x;
					found.y = c.passage[i]->p.y;
					break;
				}
			}
			found.x += (dir-1)%3-1;
			found.y += 1-(dir-1)/3;
		} while(false);
	}
	if(endpoint!=NULL) *endpoint = found;
	return true;
}
